"""Configuration file for the Sphinx documentation builder."""

import os
import sys

sys.path.insert(0, os.path.abspath("../"))

from stopuhr import __version__

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "stopuhr"
copyright = "2025, Tobias Hoelzer"
author = "Tobias Hoelzer"
version = __version__
release = __version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.coverage",
    "sphinx.ext.githubpages",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx.ext.autosummary",
    "numpydoc",
    "sphinx_copybutton",
    "sphinx_design",
    "sphinx_gallery.load_style",
    "sphinx_gallery.gen_gallery",
]

autodoc_default_options = {
    "members": True,
    "undoc-members": True,
    "show-inheritance": True,
}

# Sphinx-Gallery configuration
sphinx_gallery_conf = {
    "examples_dirs": ["../examples"],  # Path to your examples folder
    "gallery_dirs": ["auto_examples"],  # Where to save gallery files
    "backreferences_dir": "gen_modules/backreferences",
    "reference_url": {"sphinx-gallery": None},
    "plot_gallery": "True",
    "min_reported_time": 100,  # Time threshold for slow examples
    "filename_pattern": r"\.py",
}

numpydoc_show_class_members = False
templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "pydata_sphinx_theme"
html_static_path = ["_static"]
html_favicon = "_static/favicon.ico"
html_theme_options = {
    "logo": {
        "text": "stopuhr",
        "image_light": "_static/images/stopuhr-logo.png",
        "image_dark": "_static/images/stopuhr-logo.png",
    },
    "icon_links": [
        {
            # Label for this link
            "name": "GitHub",
            # URL where the link will redirect
            "url": "https://github.com/relativityhd/stopuhr",  # required
            # Icon class (if "type": "fontawesome"), or path to local image (if "type": "local")
            "icon": "fa-brands fa-square-github",
            # The type of image to be used (see below for details)
            "type": "fontawesome",
        }
    ],
    "icon_links_label": "Quick Links",
}



# ruff: noqa: D205, D212, D400, D415, E402, D103
"""
===================
Basic Usage Example
===================

This is a basic usage example of the `stopuhr` package.

"""

# %%
# Stop the time with a simple context manager and print the duration.
import time

import stopuhr

timer = stopuhr.Chronometer()

with timer("Sleeping"):
    time.sleep(0.1)

# %%
# stopuhr also provides a default instance of the `Chronometer` class, which is called `stopwatch`.
from stopuhr import stopwatch

with stopwatch("Sleeping"):
    time.sleep(0.1)

# %%
# Instead of printing, one can pass any callable to the `printer` argument, e.g. a logger.
import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s"))
logger.addHandler(handler)

with stopwatch("Sleeping", printer=logger.info):
    time.sleep(0.1)

# %%
# By default, the output is rounded to two decimal places. This can be changed with the `res` argument.
with stopwatch("Sleeping", res=3):
    time.sleep(0.16189)

# %%
# We can change the default behaviour by passing the arguments to the `__init__` method of the `Chronometer` class.
timer = stopuhr.Chronometer(printer=logger.info, res=3)
with timer("Sleeping"):
    time.sleep(0.16189)

# %%
# Since `stopwatch` is just a `Chronometer` instance, we can just overwrite the default behaviour.
stopwatch.res = 3
with stopwatch("Sleeping"):
    time.sleep(0.16189)
stopwatch.res = 2  # back to default

# %%
# The `Chronometer` class is stateful, meaning that it can be used to measure multiple durations at once.
# This also supports the `printer` and `res` arguments.
# The `log` argument of each call can be used to suppress the output.

for i in range(5):
    with stopwatch("Sleeping", log=False):
        time.sleep(0.2)

# Print a summary with the mean and standard deviation of the durations.
stopwatch.summary()

# %%
# This summary used also the previous calls to the `stopwatch` function to calculate the mean.
# We can reset it's state with the `reset` method, note that this is happening in-place.
stopwatch.reset()

# %%
# The previous behavior can also be achieved with the `start` and `stop` methods.
# Here, the `stop` method also supports the `printer`, `log` and `res` arguments.
for i in range(5):
    stopwatch.start("Sleeping")
    time.sleep(0.2)
    stopwatch.stop("Sleeping", log=False)

stopwatch.summary()

# %%
# The Chronometer can also measure multiple durations at once, thanks to that feature.

stopwatch.reset()

# Single duration
with stopwatch("A (single 0.2s sleep)", log=False):
    time.sleep(0.2)

for i in range(5):
    with stopwatch("B (multiple 0.2s sleeps)", log=False):
        time.sleep(0.2)
    with stopwatch("C (multiple 0.1s sleeps)", log=False):
        time.sleep(0.1)

stopwatch.summary()

# %%
# Because the `Chronometer`'s `__call__` method is a `@contextmanager`, it can be used as a decorator.
# This takes the same arguments as the context manager (`res`, `log` and `printer`).

stopwatch.reset()


@stopwatch("Busy Function")
def busy_function():
    time.sleep(0.2)


busy_function()

# %%
# Of course, this can be used to measure multiple functions at once.

stopwatch.reset()


@stopwatch("Busy Function", log=False)
def busy_function():
    time.sleep(0.1)


for i in range(5):
    busy_function()

stopwatch.summary()



"""Very high level benchmarking decorator."""

from collections.abc import Callable
from inspect import Signature, signature
from typing import TYPE_CHECKING, Any, Literal
from warnings import deprecated

from stopuhr.stopuhr import StopUhr, stopuhr

if TYPE_CHECKING:
    import pandas as pd


def _get_bound_args(sig: Signature, *args, **kwargs) -> dict[str, Any]:
    # Create a dictionary to store parameter bindings
    bound_args = {}

    # Bind positional arguments
    for arg_name, arg_value in zip(sig.parameters.keys(), args):
        bound_args[arg_name] = arg_value

    # Add keyword arguments
    for key, value in kwargs.items():
        if key in sig.parameters:
            bound_args[key] = value

    # Fill in default values for missing parameters
    for param_name, param in sig.parameters.items():
        if param_name not in bound_args and param.default != param.empty:
            bound_args[param_name] = param.default

    return bound_args


@deprecated("FunkUhr is deprecated. Use Chronometer instead.")
class FunkUhr:
    """Very high level benchmarking decorator.

    Wraps the `StopUhr` class to provide a decorator for benchmarking functions.

    Example:
        A stateful decorator exists as well, which is just a wrapper around the stateful timer.
        It supports the same arguments as the stateful timer.

        .. code-block:: python

            >>> from stopuhr import FunkUhr

            >>> funkuhr = FunkUhr()

            >>> @funkuhr("Busy Function", log=False)
            >>> def busy_function():
            >>>     time.sleep(0.1)

            >>> for i in range(5):
            >>>     busy_function()

            >>> funkuhr.summary()
            Busy Function took 0.10 ± 0.00 s (n=5 -> total=0.50s)

        Like the stateless decorator, it is possible to add arguments to the message.

        .. code-block:: python

            >>> from stopuhr import FunkUhr

            >>> funkuhr = FunkUhr()

            >>> @funkuhr("Busy Function", print_kwargs=["arg1", "arg2"])
            >>> def busy_function(arg1, arg2, arg3):
            >>>     time.sleep(0.1)

            >>> for i in range(5):
            >>>     busy_function(1, 2, 3)

            >>> funkuhr.summary()
            Busy Function took 0.10 ± 0.00 s (n=5 -> total=0.50s) (with arg1=1, arg2=2)

        It is also possible to add all arguments to the message.

        .. code-block:: python

            >>> from stopuhr import FunkUhr

            >>> funkuhr = FunkUhr()

            >>> @funkuhr("Busy Function", print_kwargs=True)
            >>> def busy_function(arg1, arg2):
            >>>     time.sleep(0.1)

            >>> for i in range(5):
            >>>     busy_function(1, 2)

            >>> funkuhr.summary()
            Busy Function took 0.10 ± 0.00 s (n=5 -> total=0.50s) (with arg1=1, arg2=2)

    """

    def __init__(self, printer: callable = print):
        """FunkUhr: a very high level benchmarking decorator.

        Args:
            printer (callable, optional): The function to print with. Defaults to print.

        """
        self.stopuhr = StopUhr(printer)

    def export(self) -> "pd.DataFrame":
        """Export the durations as a pandas DataFrame.

        Returns:
            pd.DataFrame: A pandas DataFrame with the durations.

        """
        return self.stopuhr.export()

    def summary(self, res: int = 2):
        """Print a summary of the durations.

        Args:
            res (int, optional): The number of decimal places to round to. Defaults to 2.

        """
        self.stopuhr.summary(res)

    def __call__(
        self,
        key: str,
        res: int = 2,
        log: bool = True,
        print_kwargs: list[str] | Literal["all"] | None = None,
    ):
        """Decorate a function to measure the time taken in a block.

        Args:
            key (str): The key to store the duration under.
            res (int, optional): The number of decimal places to round to. Defaults to 2.
            log (bool, optional): Whether to log the duration. Defaults to True.
            print_kwargs (list[str] | bool, optional): The arguments to be added to the `key`.
                If a list, only the arguments in the list will be added to the message.
                If True, all arguments will added. If False, no arguments will be added.
                Additions to the message will have the form: f"{key} (with {arg1=val1, arg2=val2, ...})".
                Defaults to False.

        Raises:
            ValueError: If any of the print_kwargs are not in the functions signature.

        """

        def _decorator(func):
            func_signature = signature(func)
            # Check if any of the print_kwargs are not in the functions signature and raise an error if so
            if isinstance(print_kwargs, list) and any(
                k not in func_signature.parameters for k in print_kwargs if isinstance(print_kwargs, list)
            ):
                raise ValueError(
                    f"Not all {print_kwargs=} found in {func_signature.parameters=} of function {func.__name__}"
                )

            def _inner(*args, **kwargs):
                _inner_key = key

                if isinstance(print_kwargs, list):
                    bound_args = _get_bound_args(func_signature, *args, **kwargs)
                    # Check if any of the print_kwargs are not in bound_args and raise an error if so
                    if any(k not in bound_args for k in print_kwargs):
                        raise ValueError(f"Not all {print_kwargs=} found in {bound_args=} of function {func.__name__}")

                    # Filter by print_kwargs
                    bound_args = {k: bound_args[k] for k in print_kwargs if k in bound_args}
                    # Make a string of it and add it to the message
                    bound_args_msg = ", ".join(f"{k}={v}" for k, v in bound_args.items())
                    _inner_key += f" (with {bound_args_msg})"
                elif print_kwargs:
                    bound_args = _get_bound_args(func_signature, *args, **kwargs)
                    # Make a string of it and add it to the message
                    bound_args_msg = ", ".join(f"{k}={v}" for k, v in bound_args.items())
                    _inner_key += f" (with {bound_args_msg})"

                with self.stopuhr(_inner_key, res=res, log=log):
                    return func(*args, **kwargs)

            return _inner

        return _decorator


@deprecated("funkuhr is deprecated. Use stopwatch instead.")
def funkuhr(msg: str, printer: callable = print, res: int = 2, print_kwargs: list[str] | bool = False):
    """Decorate a function to measure the time taken in a block.

    Wraps the `stopuhr` function to provide a decorator for benchmarking functions.

    Example:
        .. code-block:: python

            >>> from stopuhr import funkuhr

            >>> @funkuhr("Busy Function")
            >>> def busy_function():
            >>>     time.sleep(0.2)

            >>> busy_function()
            Busy Function took 0.20s

        It is possible to add arguments to the message.

        .. code-block:: python

            >>> from stopuhr import funkuhr

            >>> @funkuhr("Busy Function", print_kwargs=["arg1", "arg2"])
            >>> def busy_function(arg1, arg2, arg3):
            >>>     time.sleep(0.2)

            >>> busy_function(1, 2, 3)
            Busy Function took 0.20s (with arg1=1, arg2=2)

        It is also possible to add all arguments to the message.

        .. code-block:: python

            >>> from stopuhr import funkuhr

            >>> @funkuhr("Busy Function", print_kwargs=True)
            >>> def busy_function(arg1, arg2):
            >>>     time.sleep(0.2)

            >>> busy_function(1, 2)
            Busy Function took 0.20s (with arg1=1, arg2=2)

    Args:
        func (callable): The function to decorate.
        msg (str): The message to print.
        printer (callable, optional): The function to print with. Defaults to print.
        res (int, optional): The number of decimal places to round to. Defaults to 2.
        print_kwargs (list[str] | bool, optional): The arguments to be added to the `msg`.
            If a list, only the arguments in the list will be added to the message.
            If True, all arguments will added. If False, no arguments will be added.
            Additions to the message will have the form: f"{msg} (with {arg1=val1, arg2=val2, ...})".
            Defaults to False.

    Raises:
        ValueError: If any of the print_kwargs are not in the functions signature.

    """

    def _decorator(func: Callable):
        func_signature = signature(func)
        # Check if any of the print_kwargs are not in the functions signature and raise an error if so
        if isinstance(print_kwargs, list) and any(
            k not in func_signature.parameters for k in print_kwargs if isinstance(print_kwargs, list)
        ):
            raise ValueError(
                f"Not all {print_kwargs=} found in {func_signature.parameters=} of function {func.__name__}"
            )

        def _inner(*args, **kwargs):
            _inner_msg = msg

            if isinstance(print_kwargs, list):
                bound_args = _get_bound_args(func_signature, *args, **kwargs)
                # Filter by print_kwargs
                bound_args = {k: bound_args[k] for k in print_kwargs if k in bound_args}
                # Make a string of it and add it to the message
                bound_args_msg = ", ".join(f"{k}={v}" for k, v in bound_args.items())
                _inner_msg += f" (with {bound_args_msg})"
            elif print_kwargs:  # True was passed
                bound_args = _get_bound_args(func_signature, *args, **kwargs)
                # Make a string of it and add it to the message
                bound_args_msg = ", ".join(f"{k}={v}" for k, v in bound_args.items())
                _inner_msg += f" (with {bound_args_msg})"

            with stopuhr(_inner_msg, printer, res):
                return func(*args, **kwargs)

        return _inner

    return _decorator


__all__ = ["FunkUhr", "funkuhr"]



"""Very high level benchmarking contextmanager."""

import time
from collections import defaultdict
from contextlib import contextmanager
from statistics import mean, stdev
from typing import TYPE_CHECKING
from warnings import deprecated

if TYPE_CHECKING:
    import pandas as pd


@deprecated("StopUhr is deprecated. Use Chronometer instead.")
class StopUhr:
    """Very high level benchmarking contextmanager.

    Example:
        Use a stateful timer to measure the time taken in a loop.
        This also supports the `printer` and `res` arguments.
        The `log` argument of each call can be used to suppress the output.

        .. code-block:: python

            >>> from stopuhr import StopUhr

            >>> stopuhr = StopUhr()

            >>> for i in range(5):
            >>>     with stopuhr("Sleeping", log=False):
            >>>         time.sleep(0.2)

            >>> # Print a summary with the mean and standard deviation of the durations.
            >>> stopuhr.summary()
            Sleeping took 0.20 ± 0.00 s (n=5 -> total=1.00s)

        The `reset`` command resets the state of the timer, note that this function is happening in-place.

        .. code-block:: python

            >>> stopuhr.reset()

        The previous behavior can be achieved with the `start` and `stop` methods.
        Here, the `stop` method also supports the `log` and `res` arguments.

        .. code-block:: python

            >>> for i in range(5):
            >>>     stopuhr.start("Sleeping")
            >>>     time.sleep(0.2)
            >>>     stopuhr.stop("Sleeping", log=False)

            >>> stopuhr.summary()
            Sleeping took 0.20 ± 0.00 s (n=5 -> total=1.00s)


        The stateful timer can also measure multiple durations at once.

        .. code-block:: python

            >>> stopuhr.reset()

            >>> # Single duration
            >>> with stopuhr("A (single 0.2s sleep)", log=False):
            >>>     time.sleep(0.2)

            >>> for i in range(5):
            >>>     with stopuhr("B (multiple 0.2s sleeps)", log=False):
            >>>         time.sleep(0.2)
            >>>     with stopuhr("C (multiple 0.1s sleeps)", log=False):
            >>>         time.sleep(0.1)

            >>> stopuhr.summary()
            A (single 0.2s sleep) took 0.20 s
            B (multiple 0.2s sleeps) took 0.20 ± 0.00 s (n=5 -> total=1.00s)
            C (multiple 0.1s sleeps) took 0.10 ± 0.00 s (n=5 -> total=0.50s)

    """

    def __init__(self, printer: callable = print):
        """StopUhr: a very high level benchmarking contextmanager.

        Args:
            printer (callable, optional): The function to print with. Defaults to print.

        """
        self.printer = printer
        self.reset()

    def reset(self):
        """Reset the durations."""
        self.durations: dict[str, list[float]] = defaultdict(list)
        self.idling_starts: dict[str, list[float]] = defaultdict(list)

    def export(self) -> "pd.DataFrame":
        """Export the durations as a pandas DataFrame.

        Returns:
            pd.DataFrame: A pandas DataFrame with the durations.

        """
        import pandas as pd

        # bring durations to same length
        durations = {}
        max_len = max(len(v) for v in self.durations.values())
        for key, values in self.durations.items():
            durations[key] = values + [pd.NA] * (max_len - len(values))

        return pd.DataFrame(durations)

    def summary(self, res: int = 2):
        """Print a summary of the durations.

        Args:
            res (int, optional): The number of decimal places to round to. Defaults to 2.

        """
        for key, values in self.durations.items():
            if not values:
                self.printer(f"{key} has no durations recorded")
                continue

            if len(values) == 1:
                self.printer(f"{key} took {values[0]:.{res}f}s")
                continue

            n = len(values)
            total = sum(values)
            mean_val = mean(values)
            stdev_val = stdev(values)
            self.printer(f"{key} took {mean_val:.{res}f} ± {stdev_val:.{res}f}s (n={n} -> {total=:.{res}f}s)")

    def start(self, msg: str):
        """Start the timer for a msg.

        Args:
            msg (str): The msg / key to store the start time under.

        """
        self.idling_starts[msg].append(time.perf_counter())

    def stop(self, key: str, res: int = 2, log: bool = True):
        """Stop the timer for a key.

        Args:
            key (str): The key to store the duration under.
            res (int, optional): The number of decimal places to round to. Defaults to 2.
            log (bool, optional): Whether to log the duration. Defaults to True.

        """
        end = time.perf_counter()
        try:
            start = self.idling_starts[key].pop(0)
        except IndexError as e:
            raise ValueError(f"Key '{key}' not started!") from e
        duration = end - start
        self.durations[key].append(duration)
        if log:
            self.printer(f"{key} took {duration:.{res}f}s")

    @contextmanager
    def __call__(self, key: str, res: int = 2, log: bool = True):
        """Context manager to measure the time taken in a block.

        Args:
            key (str): The key to store the duration under.
            res (int, optional): The number of decimal places to round to. Defaults to 2.
            log (bool, optional): Whether to log the duration. Defaults to True.

        """
        start = time.perf_counter()
        yield
        duration = time.perf_counter() - start
        self.durations[key].append(duration)
        if log:
            self.printer(f"{key} took {duration:.{res}f}s")


@contextmanager
@deprecated("stopuhr is deprecated. Use stopwatch instead.")
def stopuhr(msg: str, printer: callable = print, res: int = 2):
    """Context manager to measure the time taken in a block.

    Example:
        Stop the time with a simple context manager and print the duration.

        .. code-block:: python

            >>> import time

            >>> from stopuhr import stopuhr

            >>> with stopuhr("Sleeping"):
            >>>     time.sleep(0.1)
            Sleeping took 0.10s

        Instead of printing, one can pass any callable to the `printer` argument, e.g. a logger.

        .. code-block:: python

            >>> import logging

            >>> logger = logging.getLogger(__name__)
            >>> logger.setLevel(logging.INFO)
            >>> handler = logging.StreamHandler()
            >>> handler.setFormatter(logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s"))
            >>> logger.addHandler(handler)

            >>> with stopuhr("Sleeping", printer=logger.info):
            >>>     time.sleep(0.1)
            2025-03-27 19:11:10,912 - __main__ - INFO - Sleeping took 0.10s

        By default, the output is rounded to two decimal places. This can be changed with the `res` argument.

        .. code-block:: python

            >>> with stopuhr("Sleeping", res=3):
            >>>     time.sleep(0.16189)
            Sleeping took 0.162 s

    Args:
        msg (str): The message to print.
        printer (callable, optional): The function to print with. Defaults to print.
        res (int, optional): The number of decimal places to round to. Defaults to 2.

    """
    start = time.perf_counter()
    yield
    duration = time.perf_counter() - start
    printer(f"{msg} took {duration:.{res}f}s")


__all__ = ["StopUhr", "stopuhr"]



# ruff: noqa: D205, D212, D400, D415, E402, D103
"""
======================
Advanced Usage Example
======================

This is a advanced usage example of the `stopuhr` package.

"""

# %%
# Also print out function arguments
import time

from stopuhr import stopwatch


@stopwatch.f("Sleeping")
def sleep(seconds: float):
    time.sleep(seconds)


sleep(0.1)

# %%
# If some parameters should not be printed,
# we can use the `print_kwargs` argument.to select which ones should be printed.


@stopwatch.f("Sleeping", print_kwargs=["seconds"])
def sleep(seconds: float, other: str):
    time.sleep(seconds)
    print(other)


sleep(0.1, "Hello")

# %%
# In multi processing situations, we must use a separate `Chronometer` instance for each process.
# For this, we can use the `merge` or `combine` method to combine the results of multiple `Chronometer` instances.

import stopuhr


def simulate_process():
    timer = stopuhr.Chronometer()

    with timer("Sleeping"):
        time.sleep(0.1)
        res = 3.14159
    return {"result": res, "timer": timer}


# We only simulate a multi processing situation here. A real one would be too much overhead.

tasks_outputs = [simulate_process() for _ in range(5)]
tasks_timer = [task["timer"] for task in tasks_outputs]

combined_timer = stopuhr.Chronometer.combine(tasks_timer)
combined_timer.summary()



"""Very high level benchmarking contextmanager."""

import time
from collections import defaultdict
from collections.abc import Callable
from contextlib import contextmanager
from inspect import Signature, signature
from statistics import mean, stdev
from typing import TYPE_CHECKING, Any, Literal

if TYPE_CHECKING:
    import pandas as pd


# Pritner signature
Printer = Callable[[str], None]


def _get_bound_args(sig: Signature, *args, **kwargs) -> dict[str, Any]:
    # Create a dictionary to store parameter bindings
    bound_args = {}

    # Bind positional arguments
    for arg_name, arg_value in zip(sig.parameters.keys(), args):
        bound_args[arg_name] = arg_value

    # Add keyword arguments
    for key, value in kwargs.items():
        if key in sig.parameters:
            bound_args[key] = value

    # Fill in default values for missing parameters
    for param_name, param in sig.parameters.items():
        if param_name not in bound_args and param.default != param.empty:
            bound_args[param_name] = param.default

    return bound_args


class Chronometer:
    """Very high level benchmarking contextmanager and decorator.

    Example:
        Measure the time taken in a code block.

        .. code-block:: python

            >>> import stopuhr

            >>> timer = stopuhr.Chronometer()
            >>> with timer("Sleeping"):
            >>>     time.sleep(0.2)

        Instead of printing, one can pass any callable to the `printer` argument, e.g. a logger.

        .. code-block:: python

            >>> import logging

            >>> logger = logging.getLogger(__name__)
            >>> logger.setLevel(logging.INFO)
            >>> handler = logging.StreamHandler()
            >>> handler.setFormatter(logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s"))
            >>> logger.addHandler(handler)

            >>> timer = stopuhr.Chronometer(printer=logger.info)
            >>> with timer("Sleeping"):
            >>>     time.sleep(0.1)
            2025-03-27 19:11:10,912 - __main__ - INFO - Sleeping took 0.10s

        The printer can be overridden for each call.

        .. code-block:: python

            >>> timer = stopuhr.Chronometer(printer=logger.info)
            >>> with timer("Sleeping", printer=logger.debug):
            >>>     time.sleep(0.1)
            2025-03-27 19:11:10,912 - __main__ - DEBUG - Sleeping took 0.10s

        By default, the output is rounded to two decimal places. This can be changed with the `res` argument.
        The `res` argument can also be overridden for each call.

        .. code-block:: python

            >>> timer = stopuhr.Chronometer(res=3)
            >>> with timer("Sleeping"):
            >>>     time.sleep(0.16189)
            Sleeping took 0.162 s

            >>> with timer("Sleeping", res=2):
            >>>     time.sleep(0.16189)
            Sleeping took 0.16 s

        The chronometer is stateful and can be used to measure functions inside a loop.
        The `log` argument of each call can be used to suppress the output.
        It can also be set individually for each call.

        .. code-block:: python

            >>> timer = stopuhr.Chronometer(log=False)

            >>> for i in range(5):
            >>>     with timer("Sleeping"):
            >>>         time.sleep(0.2)

            >>> # Print a summary with the mean and standard deviation of the durations.
            >>> timer.summary()
            Sleeping took 0.20 ± 0.00 s (n=5 -> total=1.00s)

        The `reset` command resets the state of the timer, note that this function is happening in-place.

        .. code-block:: python

            >>> timer.reset()

        The previous behavior can be achieved with the `start` and `stop` methods.
        Here, the `stop` method also supports the `log` and `res` arguments.

        .. code-block:: python

            >>> for i in range(5):
            >>>     timer.start("Sleeping")
            >>>     time.sleep(0.2)
            >>>     timer.stop("Sleeping", log=False)

            >>> timer.summary()
            Sleeping took 0.20 ± 0.00 s (n=5 -> total=1.00s)


        The stateful timer can also measure multiple durations at once.

        .. code-block:: python

            >>> timer.reset()

            >>> # Single duration
            >>> with timer("A (single 0.2s sleep)", log=False):
            >>>     time.sleep(0.2)

            >>> for i in range(5):
            >>>     with timer("B (multiple 0.2s sleeps)", log=False):
            >>>         time.sleep(0.2)
            >>>     with timer("C (multiple 0.1s sleeps)", log=False):
            >>>         time.sleep(0.1)

            >>> timer.summary()
            A (single 0.2s sleep) took 0.20 s
            B (multiple 0.2s sleeps) took 0.20 ± 0.00 s (n=5 -> total=1.00s)
            C (multiple 0.1s sleeps) took 0.10 ± 0.00 s (n=5 -> total=0.50s)

        The `export` method can be used to export the durations as a pandas DataFrame.

        .. code-block:: python

            >>> import pandas as pd

            >>> timer.reset()

            >>> for i in range(5):
            >>>     with timer("Sleeping", log=False):
            >>>         time.sleep(0.2)
            >>> for i in range(3):
            >>>     with timer("Sleeping-2", log=False):
            >>>         time.sleep(0.2)

            >>> df = timer.export()
            >>> print(df)
              Sleeping-2  Sleeping
            0  0.2         0.2
            1  0.2         0.2
            2  0.2         0.2
            3  0.2
            4  0.2

    """

    def __init__(self, printer: Printer = print, res: int = 2, log: bool = True):
        """StopUhr: a very high level benchmarking contextmanager.

        Args:
            printer (Callable[[str], None], optional): The objects default function to print to.
                This can be a logger function, e.g. `logger.info`, the normal `print` statement,
                or any other function which receives exactly one string as input.
                Output methods may overwrite this default.
                Defaults to print.
            res (int, optional): The default number of decimal places to round to in the output.
                Output methods may overwrite this default.
                Defaults to 2.
            log (bool, optional): Whether to log the duration at occurence.
                Setting this to `False` will not output anything,
                it will only add the message and the duration to the objects state as usual.
                Output methods may overwrite this default.
                Defaults to True.

        """
        self.printer = printer
        self.res = res
        self.log = log
        self.reset()

    def reset(self):
        """Reset the durations."""
        self.durations: dict[str, list[float]] = defaultdict(list)
        self.idling_starts: dict[str, list[float]] = defaultdict(list)

    def export(self) -> "pd.DataFrame":
        """Export the durations as a pandas DataFrame.

        Returns:
            pd.DataFrame: A pandas DataFrame with the durations.

        """
        import pandas as pd

        # bring durations to same length
        durations = {}
        max_len = max(len(v) for v in self.durations.values())
        for key, values in self.durations.items():
            durations[key] = values + [pd.NA] * (max_len - len(values))

        return pd.DataFrame(durations)

    def merge(self, other: "Chronometer"):
        """Merge the durations of another Chronometer into this one.

        This is handy for multiprocessing situations.

        Does not merge the idling_starts, only the durations.

        Args:
            other (Chronometer): The Chronometer to merge with.
                This Chronometer will be modified in-place.
                The other Chronometer will not be modified.

        """
        assert isinstance(other, Chronometer), "other must be a Chronometer"

        for key, values in other.durations.items():
            self.durations[key].extend(values)

    @staticmethod
    def combine(*timers: "Chronometer") -> "Chronometer":
        """Combine multiple Chronometers into one.

        This is handy for multiprocessing situations.

        Args:
            *timers (Chronometer): The Chronometers to combine.
                The new Chronometer will be created with the same printer, res and log settings as the first one.

        Returns:
            Chronometer: A new Chronometer with the combined durations.

        """
        # Check if the first input is a list of Chronometers
        if isinstance(timers[0], list):
            timers = timers[0]
        assert all(isinstance(t, Chronometer) for t in timers), "all timers must be Chronometers"
        assert len(timers) > 0, "at least one timer must be provided"

        combined = Chronometer(printer=timers[0].printer, res=timers[0].res, log=timers[0].log)
        for timer in timers:
            combined.merge(timer)

        return combined

    def summary(self, res: int | None = None, printer: Printer | None = None):
        """Print a summary of the durations.

        Args:
            res (int | None, optional): Override the objects default number of decimal places to round to if not None.
                Defaults to None.
            printer (Callable[[str], None] | None, optional): Override the objects function to print with if not None.
                 Defaults to None.

        """
        res = res or self.res
        printer = printer or self.printer
        for key, values in self.durations.items():
            if not values:
                printer(f"{key} has no durations recorded")
                continue

            if len(values) == 1:
                printer(f"{key} took {values[0]:.{res}f}s")
                continue

            n = len(values)
            total = sum(values)
            mean_val = mean(values)
            stdev_val = stdev(values)
            printer(f"{key} took {mean_val:.{res}f} ± {stdev_val:.{res}f}s (n={n} -> {total=:.{res}f}s)")

    def start(self, key: str):
        """Start the timer for a msg.

        Args:
            key (str): The msg / key to store the start time under.

        """
        self.idling_starts[key].append(time.perf_counter())

    def stop(self, key: str, res: int | None = None, log: bool | None = None, printer: Printer | None = None):
        """Stop the timer for a key.

        Args:
            key (str): The key to store the duration under.
            res (int | None, optional): Override the objects default number of decimal places to round to if not None.
                Defaults to None.
            log (bool | None, optional): Override the objects default log setting if not None.
                Defaults to None.
            printer (Callable[[str], None] | None, optional): Override the objects function to print with if not None.
                 Defaults to None.

        """
        res = res or self.res
        log = log if log is not None else self.log
        printer = printer or self.printer
        end = time.perf_counter()
        try:
            start = self.idling_starts[key].pop(0)
        except IndexError as e:
            raise ValueError(f"Key '{key}' not started!") from e
        duration = end - start
        self.durations[key].append(duration)
        if log:
            printer(f"{key} took {duration:.{res}f}s")

    def f(
        self,
        key: str,
        res: int | None = None,
        log: bool | None = None,
        printer: Printer | None = None,
        print_kwargs: list[str] | Literal["all"] | None = "all",
    ):
        """Advanced decorator to also print the function arguments.

        Instead of just passing the normal `{key} took {duration:.{res}f}s` message,
        this decorator will also add the function arguments to the message.

        This behaviour is not possible to implement with the `@contextmanager` decorator,
        hence the need for this separate decorator.

        Args:
            key (str): The key to store the duration under.
            res (int | None, optional): Override the objects default number of decimal places to round to if not None.
                Defaults to None.
            log (bool | None, optional): Override the objects default log setting if not None.
                Defaults to None.
            printer (Callable[[str], None] | None, optional): Override the objects function to print with if not None.
                 Defaults to None.
            print_kwargs (list[str] | bool, optional): The arguments to be added to the `key`.
                If a list, only the arguments in the list will be added to the message.
                If True, all arguments will added. If False, no arguments will be added.
                Additions to the message will have the form: f"{key} (with {arg1=val1, arg2=val2, ...})".
                Defaults to False.

        Raises:
            ValueError: If any of the print_kwargs are not in the functions signature.

        """
        res = res or self.res
        log = log if log is not None else self.log
        printer = printer or self.printer

        def _decorator(func):
            func_signature = signature(func)
            # Check if any of the print_kwargs are not in the functions signature and raise an error if so
            if isinstance(print_kwargs, list) and any(
                k not in func_signature.parameters for k in print_kwargs if isinstance(print_kwargs, list)
            ):
                raise ValueError(
                    f"Not all {print_kwargs=} found in {func_signature.parameters=} of function {func.__name__}"
                )

            def _inner(*args, **kwargs):
                _inner_key = key

                if isinstance(print_kwargs, list):
                    bound_args = _get_bound_args(func_signature, *args, **kwargs)
                    # Check if any of the print_kwargs are not in bound_args and raise an error if so
                    if any(k not in bound_args for k in print_kwargs):
                        raise ValueError(f"Not all {print_kwargs=} found in {bound_args=} of function {func.__name__}")

                    # Filter by print_kwargs
                    bound_args = {k: bound_args[k] for k in print_kwargs if k in bound_args}
                    # Make a string of it and add it to the message
                    bound_args_msg = ", ".join(f"{k}={v}" for k, v in bound_args.items())
                    _inner_key += f" (with {bound_args_msg})"
                elif print_kwargs:
                    bound_args = _get_bound_args(func_signature, *args, **kwargs)
                    # Make a string of it and add it to the message
                    bound_args_msg = ", ".join(f"{k}={v}" for k, v in bound_args.items())
                    _inner_key += f" (with {bound_args_msg})"

                with self(_inner_key, res=res, log=log):
                    return func(*args, **kwargs)

            return _inner

        return _decorator

    @contextmanager
    def __call__(
        self,
        key: str,
        res: int | None = None,
        log: bool | None = None,
        printer: Printer | None = None,
    ):
        """Context manager and decorator to measure the time taken in a block or function.

        Args:
            key (str): The key to store the duration under.
            res (int | None, optional): Override the objects default number of decimal places to round to if not None.
                Defaults to None.
            log (bool | None, optional): Override the objects default log setting if not None.
                Defaults to None.
            printer (Callable[[str], None] | None, optional): Override the objects function to print with if not None.
                 Defaults to None.

        """
        self.start(key)
        yield
        self.stop(key, res=res, log=log, printer=printer)


stopwatch = Chronometer()

__all__ = ["Chronometer", "stopwatch"]



import time

from stopuhr import Chronometer, stopwatch


def assert_out(excpected_start: str):
    def _innter(msg: str):
        assert msg.startswith(excpected_start), f"Expected {msg} to start with '{excpected_start}'"

    return _innter


def assert_never():
    def _inner(msg: str):
        raise AssertionError(f"Expected no output, but got: {msg}")

    return _inner


def test_default_context_manager():
    with stopwatch("test", printer=assert_out("test took")):
        time.sleep(0.1)


def test_default_decorator():
    @stopwatch("test", printer=assert_out("test took"))
    def test_function():
        time.sleep(0.1)

    test_function()


def test_default_no_log():
    with stopwatch("test", printer=assert_never(), log=False):
        time.sleep(0.1)


def test_default_res():
    with stopwatch("test", printer=assert_out("test took 0.100"), res=3):
        time.sleep(0.1)


def test_summary():
    timer = Chronometer()
    with timer("test", log=False):
        time.sleep(0.1)

    with timer("test", log=False):
        time.sleep(0.3)

    with timer("test2", log=False):
        time.sleep(0.2)

    out = []
    stopwatch.summary(printer=out.append, res=1)
    assert out[0] == "test took 0.2 ± 0.1s (n=2 -> total=0.4s)"
    assert out[1] == "test2 took 0.2s"


def test_default_export():
    timer = Chronometer()
    with timer("test", log=False):
        time.sleep(0.1)

    with timer("test", log=False):
        time.sleep(0.3)

    with timer("test2", log=False):
        time.sleep(0.2)

    times = timer.export()
    assert len(times) == 2
    assert len(times.columns) == 2
    assert times.iloc[:, 0].count() == 2
    assert times.iloc[:, 1].count() == 1
    assert abs(times.iloc[:, 0].sum() - 0.4) < 0.01
    assert abs(times.iloc[:, 1].sum() - 0.2) < 0.01


def test_adv_decorator():
    @stopwatch.f("test", printer=assert_out("test (with a=1, b=2, c=3) took"))
    def test_function(a, b, c):
        time.sleep(0.1)

    test_function(1, 2, 3)


def test_adv_decorator_with_arg_filter():
    @stopwatch.f("test", printer=assert_out("test (with a=1, c=3) took"), print_kwargs=["a", "c"])
    def test_function(a, b, c):
        time.sleep(0.1)

    test_function(1, 2, 3)



"""stopuhr: A simple tool for measuring durations in Python."""

import importlib.metadata

from stopuhr.chrono import Chronometer, stopwatch
from stopuhr.funkuhr import FunkUhr, funkuhr
from stopuhr.stopuhr import StopUhr, stopuhr

try:
    __version__ = importlib.metadata.version("stopuhr")
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.0.0"

__all__ = ["Chronometer", "FunkUhr", "StopUhr", "funkuhr", "stopuhr", "stopwatch"]
